import os
import shutil
import tempfile

import pkg_resources

from z3c.dependencychecker.package import PackageMetadata


def test_no_setup_py_file_found():
    folder = tempfile.mkdtemp()
    sys_exit = False
    try:
        PackageMetadata(folder)
    except SystemExit:
        sys_exit = True

    assert sys_exit


def test_distribution_root(minimal_structure):
    path, package_name = minimal_structure
    metadata = PackageMetadata(path)

    assert metadata.distribution_root == path


def test_setup_py_path(minimal_structure):
    path, package_name = minimal_structure
    metadata = PackageMetadata(path)

    assert metadata.setup_py_path == os.path.join(path, 'setup.py')


def test_package_dir_on_distribution_root(minimal_structure):
    path, package_name = minimal_structure
    metadata = PackageMetadata(path)

    assert metadata.package_dir == path


def test_package_dir_on_src_folder(minimal_structure):
    def move_top_level_to_src(package_path, egg_folder, src_path):
        top_level_file_path = os.path.join(egg_folder, 'top_level.txt')
        with open(top_level_file_path) as top_level_file:
            top_level_folder = top_level_file.read().strip()

        top_level_sources = os.path.join(package_path, top_level_folder)

        shutil.move(top_level_sources, src_path)
        shutil.move(egg_folder, src_path)

    path, package_name = minimal_structure

    egg_info_folder = os.path.join(path, f'{package_name}.egg-info')
    src_folder = os.path.join(path, 'src')

    move_top_level_to_src(path, egg_info_folder, src_folder)

    metadata = PackageMetadata(path)

    assert metadata.package_dir == src_folder


def test_package_dir_on_another_folder(minimal_structure):
    def move_top_level_to_folder(package_path, egg_folder, folder_path):
        top_level_file_path = os.path.join(egg_folder, 'top_level.txt')
        with open(top_level_file_path) as top_level_file:
            top_level_folder = top_level_file.read().strip()

        top_level_sources = os.path.join(package_path, top_level_folder)

        shutil.move(top_level_sources, folder_path)
        shutil.move(egg_folder, folder_path)

    path, package_name = minimal_structure

    egg_info_folder = os.path.join(path, f'{package_name}.egg-info')
    folder_path = os.path.join(path, 'somewhere')

    move_top_level_to_folder(path, egg_info_folder, folder_path)

    metadata = PackageMetadata(path)

    assert metadata.package_dir == folder_path


def test_no_package_dir_found(minimal_structure):
    path, package_name = minimal_structure
    shutil.rmtree(os.path.join(path, f'{package_name}.egg-info'))

    sys_exit = False
    try:
        PackageMetadata(path)
    except SystemExit:
        sys_exit = True

    assert sys_exit


def test_no_package_dir_and_no_src_folder(minimal_structure):
    path, package_name = minimal_structure
    shutil.rmtree(os.path.join(path, f'{package_name}.egg-info'))
    shutil.rmtree(os.path.join(path, 'src'))

    sys_exit = False
    try:
        PackageMetadata(path)
    except SystemExit:
        sys_exit = True

    assert sys_exit


def test_egg_info_dir_path(minimal_structure):
    path, package_name = minimal_structure
    metadata = PackageMetadata(path)

    egg_info_path = os.path.join(path, f'{package_name}.egg-info')

    assert metadata.egg_info_dir == egg_info_path


def test_package_name(minimal_structure):
    path, package_name = minimal_structure
    metadata = PackageMetadata(path)

    assert metadata.name == package_name


def test_working_set(minimal_structure):
    path, package_name = minimal_structure
    metadata = PackageMetadata(path)

    assert isinstance(metadata._working_set, pkg_resources.WorkingSet)


def test_package_in_working_set(minimal_structure):
    path, package_name = minimal_structure
    metadata = PackageMetadata(path)

    ourselves = metadata._get_ourselves_from_working_set()
    assert ourselves.project_name == package_name


def test_package_missing_pkg_info_file(minimal_structure):
    path, package_name = minimal_structure

    pkg_info = os.path.join(
        path,
        f'{package_name}.egg-info',
        'PKG-INFO',
    )
    os.remove(pkg_info)

    metadata = PackageMetadata(path)

    sys_exit = False
    try:
        metadata._get_ourselves_from_working_set()
    except SystemExit:
        sys_exit = True

    assert sys_exit


def test_package_broken_pkg_info_file(minimal_structure):
    path, package_name = minimal_structure

    pkg_info_path = os.path.join(
        path,
        f'{package_name}.egg-info',
        'PKG-INFO',
    )
    os.remove(pkg_info_path)
    with open(pkg_info_path, 'w') as pkg_info_file:
        pkg_info_file.write('\n'.join(['Name: bla']))

    metadata = PackageMetadata(path)
    sys_exit = False
    try:
        metadata._get_ourselves_from_working_set()
    except SystemExit:
        sys_exit = True

    assert sys_exit


def test_get_dependencies(minimal_structure):
    path, package_name = minimal_structure
    metadata = PackageMetadata(path)

    names = [x.name for x in metadata.get_required_dependencies()]
    assert 'one' in names
    assert 'two' in names


def test_get_no_dependencies(minimal_structure):
    path, package_name = minimal_structure

    requires_file_path = os.path.join(
        path,
        f'{package_name}.egg-info',
        'requires.txt',
    )
    os.remove(requires_file_path)

    metadata = PackageMetadata(path)

    requirements = list(metadata.get_required_dependencies())
    assert len(requirements) == 0


def test_dependencies_with_extras(minimal_structure):
    path, package_name = minimal_structure
    requires_file_path = os.path.join(
        path,
        f'{package_name}.egg-info',
        'requires.txt',
    )
    with open(requires_file_path, 'w') as requires:
        requires.write(
            '\n'.join(
                [
                    'one [with_extra]',
                    'another[with_extra]',
                    'yet[one,two,three,extras]',
                    'something[only, two, with, spaces]',
                ]
            )
        )

    metadata = PackageMetadata(path)
    names = [x.name for x in metadata.get_required_dependencies()]

    assert len(names) == 4
    assert 'one' in names
    assert 'another' in names
    assert 'yet' in names
    assert 'something' in names


def test_dependencies_with_version_specifiers(minimal_structure):
    path, package_name = minimal_structure
    requires_file_path = os.path.join(
        path,
        f'{package_name}.egg-info',
        'requires.txt',
    )
    with open(requires_file_path, 'w') as requires:
        requires.write(
            '\n'.join(
                ['one > 8.5', 'another<=3.4', 'yet==5.2', 'something >=2.2.1,<2.3dev']
            )
        )

    metadata = PackageMetadata(path)
    names = [x.name for x in metadata.get_required_dependencies()]

    assert len(names) == 4
    assert 'one' in names
    assert 'another' in names
    assert 'yet' in names


def test_get_extra_dependencies(minimal_structure):
    path, package_name = minimal_structure

    requires_file_path = os.path.join(
        path,
        f'{package_name}.egg-info',
        'requires.txt',
    )
    with open(requires_file_path, 'w') as requires_file:
        requires_file.write('\n'.join(['one', '', '[test]', 'pytest', 'mock']))

    metadata = PackageMetadata(path)
    extras = list(metadata.get_extras_dependencies())
    extra_packages = [x.name for x in extras[0][1]]
    assert len(extras) == 1
    assert 'pytest' in extra_packages
    assert 'mock' in extra_packages


def test_no_extras(minimal_structure):
    path, package_name = minimal_structure

    metadata = PackageMetadata(path)

    extras = list(metadata.get_extras_dependencies())
    assert len(extras) == 0


def test_top_level_txt_file_found(minimal_structure):
    path, package_name = minimal_structure
    metadata = PackageMetadata(path)

    assert metadata.top_level == [os.path.join(path, package_name)]


def test_no_top_level_txt_file_found(minimal_structure):
    path, package_name = minimal_structure
    os.remove(os.path.join(path, f'{package_name}.egg-info', 'top_level.txt'))

    sys_exit = False
    metadata = PackageMetadata(path)
    try:
        metadata.top_level
    except SystemExit:
        sys_exit = True

    assert sys_exit


def test_no_sources_top_level_folder_found(minimal_structure):
    path, package_name = minimal_structure
    os.removedirs(os.path.join(path, package_name))

    sys_exit = False
    metadata = PackageMetadata(path)
    try:
        metadata.top_level
    except SystemExit:
        sys_exit = True

    assert sys_exit


def test_top_level_is_module(minimal_structure):
    path, package_name = minimal_structure
    top_level_path = os.path.join(path, package_name)
    os.removedirs(top_level_path)
    top_level_module_path = f'{top_level_path}.py'
    with open(top_level_module_path, 'w') as top_level_file:
        top_level_file.write('Does not matter')

    metadata = PackageMetadata(path)

    assert metadata.top_level == [top_level_module_path]


def test_top_level_multiple(minimal_structure):
    path, package_name = minimal_structure
    top_level_file = os.path.join(
        path,
        f'{package_name}.egg-info',
        'top_level.txt',
    )
    with open(top_level_file, 'w') as top_level:
        top_level.write('one\n')
        top_level.write('two\n')
        top_level.write('three\n')

    top_level_folders = [
        f'{path}/one',
        f'{path}/two',
        f'{path}/three',
    ]
    for new_top_level in top_level_folders:
        os.makedirs(new_top_level)

    metadata = PackageMetadata(path)

    assert metadata.top_level == top_level_folders



import logging

logger = logging.getLogger(__name__)


class Report:
    def __init__(self, package):
        self._database = package.imports
        self.exit_status = 0

    def print_report(self):
        logger.debug('Package requirements: %s', self._database._requirements)
        logger.debug('Package extras: %s', self._database._extras_requirements)
        logger.debug(
            'User defined mappings: %s',
            self._database.user_mappings,
        )
        self.missing_requirements()
        self.missing_test_requirements()
        self.unneeded_requirements()
        self.requirements_that_should_be_test_requirements()
        self.unneeded_test_requirements()
        self.print_notice()

    def missing_requirements(self):
        self._print_metric(
            'Missing requirements',
            self._database.get_missing_imports,
        )

    def missing_test_requirements(self):
        self._print_metric(
            'Missing test requirements',
            self._database.get_missing_test_imports,
        )

    def unneeded_requirements(self):
        self._print_metric(
            'Unneeded requirements',
            self._database.get_unneeded_requirements,
        )

    def requirements_that_should_be_test_requirements(self):
        self._print_metric(
            'Requirements that should be test requirements',
            self._database.requirements_that_should_be_test_requirements,
        )

    def unneeded_test_requirements(self):
        self._print_metric(
            'Unneeded test requirements',
            self._database.get_unneeded_test_requirements,
        )

    @staticmethod
    def print_notice():
        print('')
        print('Note: requirements are taken from the egginfo dir, so you need')
        print('to re-run buildout (or setup.py or whatever) for changes in')
        print('setup.py to have effect.')
        print('')

    def _print_metric(self, title, method):
        missed = method()
        if len(missed) == 0:
            return

        self.exit_status = 1

        self._print_header(title)
        for dotted_name in missed:
            print(f'     {dotted_name.name}')

    @staticmethod
    def _print_header(message):
        if message:
            print('')
            print(message)
            print('=' * len(message))



import logging
import optparse
import os
import sys

import pkg_resources

from z3c.dependencychecker.package import Package
from z3c.dependencychecker.report import Report

logger = logging.getLogger(__name__)


def main():
    options, args = parse_command_line()
    set_log_level(options.verbose)
    path = get_path(args)

    package_analyzed = Package(path)
    package_analyzed.inspect()

    report = Report(package_analyzed)
    report.print_report()

    if options.exit_status:
        exit(report.exit_status)

    exit(0)


def parse_command_line():
    usage = 'Usage: %prog [path]\n(path defaults to package name, fallback is "src/")'
    parser = optparse.OptionParser(usage=usage, version=_version())
    parser.add_option(
        '-v',
        '--verbose',
        action='store_true',
        dest='verbose',
        default=False,
        help='Show debug output',
    )
    parser.add_option(
        '--exit-zero',
        action='store_false',
        dest='exit_status',
        default=True,
        help='Exit with status code "0" even if there are errors.',
    )
    options, args = parser.parse_args()
    return options, args


def _version():
    ourselves = pkg_resources.require('z3c.dependencychecker')[0]
    return ourselves.version


def set_log_level(verbose):
    level = logging.INFO
    if verbose:
        level = logging.DEBUG

    logging.basicConfig(
        level=level, stream=sys.stdout, format='%(levelname)s: %(message)s'
    )


def get_path(args):
    """Get path to the python source distribution that will be checked for
    dependencies

    If no path is given on the command line arguments, the current working
    directory is used instead.
    """
    path = os.getcwd()

    if len(args) < 1:
        logger.debug('path used: %s', path)
        return path

    path = os.path.abspath(args[0])
    if os.path.isdir(path):
        logger.debug('path used: %s', path)
        return path

    logger.error('Given path is not a folder: %s', args[0])
    sys.exit(1)



import os

from z3c.dependencychecker.package import Package, PackageMetadata
from z3c.dependencychecker.tests.utils import (
    get_extras_requirements_names,
    get_requirements_names,
    get_requirements_names_for_extra,
    get_sorted_imports_paths,
    write_source_file_at,
)


def test_package_has_metadata(minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)

    assert bool(getattr(package, 'metadata', False))
    assert isinstance(package.metadata, PackageMetadata)


def test_declared_dependencies(minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    package.set_declared_dependencies()

    requirements = get_requirements_names(package.imports)
    assert len(requirements) == 2
    assert 'one' in requirements
    assert 'two' in requirements


def test_declared_dependencies_empty(minimal_structure):
    path, package_name = minimal_structure
    requires_file_path = os.path.join(
        path,
        f'{package_name}.egg-info',
        'requires.txt',
    )
    with open(requires_file_path, 'w') as requires:
        requires.write('')

    package = Package(path)
    package.set_declared_dependencies()

    assert len(get_requirements_names(package.imports)) == 0


def test_declared_extras_dependencies_one_extra(minimal_structure):
    path, package_name = minimal_structure
    requires_file_path = os.path.join(
        path,
        f'{package_name}.egg-info',
        'requires.txt',
    )
    with open(requires_file_path, 'w') as requires:
        requires.write('\n'.join(['[extra]', 'my.package', 'another.package']))

    package = Package(path)
    package.set_declared_dependencies()
    package.set_declared_extras_dependencies()

    extras_names = get_extras_requirements_names(package.imports)
    assert len(extras_names) == 1
    assert 'extra' in extras_names
    names = get_requirements_names_for_extra(package.imports, extra='extra')
    assert 'another.package' in names
    assert 'my.package' in names


def test_declared_extras_dependencies_only_on_extras(minimal_structure):
    """Check that main dependencies are not bundled on the extras"""
    path, package_name = minimal_structure
    requires_file_path = os.path.join(
        path,
        f'{package_name}.egg-info',
        'requires.txt',
    )
    with open(requires_file_path, 'w') as requires:
        requires.write(
            '\n'.join(['setuptools', '', '[extra]', 'my_package', 'my_other_package'])
        )

    package = Package(path)
    package.set_declared_dependencies()
    package.set_declared_extras_dependencies()

    names = get_requirements_names_for_extra(package.imports, extra='extra')
    assert 'setuptools' not in names


def test_multiple_extras(minimal_structure):
    path, package_name = minimal_structure
    requires_file_path = os.path.join(
        path,
        f'{package_name}.egg-info',
        'requires.txt',
    )
    with open(requires_file_path, 'w') as requires:
        requires.write(
            '\n'.join(
                [
                    'setuptools',
                    '',
                    '[extra]',
                    'my_package',
                    'my_other_package',
                    '[second]',
                    'just',
                    'laughs',
                ]
            )
        )

    package = Package(path)
    package.set_declared_extras_dependencies()

    extras_names = get_extras_requirements_names(package.imports)
    assert len(extras_names) == 2
    assert 'second' in extras_names
    assert 'extra' in extras_names


def test_no_python_modules_found(minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    package.analyze_package()
    assert len(package.imports.imports_used) == 0


def test_python_modules_on_top_level_sources(minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at((path, package_name), '__init__.py')
    package = Package(path)
    package.analyze_package()

    assert len(package.imports.imports_used) == 1
    paths = get_sorted_imports_paths(package.imports)
    assert paths[0].endswith('__init__.py')


def test_python_modules_found_inner_folders(minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at((path, package_name), '__init__.py')
    write_source_file_at((path, package_name, 'test'), '__init__.py')
    write_source_file_at((path, package_name, 'test'), 'test.py')

    package = Package(path)
    package.analyze_package()
    paths = get_sorted_imports_paths(package.imports)
    assert len(paths) == 3
    assert paths[0].endswith('__init__.py')
    assert paths[1].endswith('__init__.py')
    assert paths[2].endswith('test.py')


def test_top_level_is_a_file(minimal_structure):
    path, package_name = minimal_structure

    egg_info_folder = os.path.join(path, f'{package_name}.egg-info')
    top_level_file_path = os.path.join(egg_info_folder, 'top_level.txt')
    with open(top_level_file_path) as top_level_file:
        top_level_folder = top_level_file.read().strip()

    top_level_sources = os.path.join(path, top_level_folder)
    os.removedirs(top_level_sources)
    with open(f'{top_level_sources}.py', 'w') as top_level_file:
        top_level_file.write('import one')

    package = Package(path)
    package.analyze_package()
    paths = get_sorted_imports_paths(package.imports)
    assert len(paths) == 1
    assert paths[0].endswith(f'{top_level_folder}.py')



import logging
import sys

from z3c.dependencychecker.dotted_name import DottedName

logger = logging.getLogger(__name__)


# starting from python 3.10 the list of builtin methods are available directly
PY_10_OR_HIGHER = sys.version_info[1] >= 10


class ImportsDatabase:
    """Store all imports and requirements of a package

    Use that information to extract useful data out of it,
    like requirements unused, etc.
    """

    def __init__(self):
        self._requirements = set()
        self._extras_requirements = {}
        self.imports_used = []
        self.user_mappings = {}
        self.reverse_user_mappings = {}
        self.ignored_packages = set()
        self.own_dotted_name = None

    def add_requirements(self, requirements):
        self._requirements = set(requirements)

    def add_extra_requirements(self, extra_name, dotted_names):
        # A bit of extra work needs to be done as pkg_resources API returns
        # all common requirements as well as the extras when asking for an
        # extra requirements.
        only_extra_dotted_names = self._filter_duplicates(dotted_names)
        if extra_name in self._extras_requirements.keys():
            self._extras_requirements[extra_name].update(
                only_extra_dotted_names,
            )

            logger.warning(
                'extra requirement "%s" is declared twice in setup.py',
                extra_name,
            )
        else:
            self._extras_requirements[extra_name] = only_extra_dotted_names

    def _filter_duplicates(self, imports):
        """Return all items in imports that are not a requirement already"""
        all_imports = set(imports)
        filtered = all_imports - self._requirements
        return filtered

    def add_imports(self, imports):
        filters = (
            self._filter_out_known_packages,
            self._filter_out_own_package,
            self._filter_out_python_standard_library,
        )
        for single_import in imports:
            logger.debug('    Import found: %s', single_import.name)

            unknown_import = self._apply_filters([single_import], filters)
            if unknown_import:
                self.imports_used.append(single_import)
            else:
                logger.debug('    Import ignored: %s', single_import.name)

    def add_user_mapping(self, package_name, provided_names):
        package = DottedName(package_name)
        packages_provided = [DottedName(name) for name in provided_names]

        if package not in self._all_requirements():
            logger.info(
                'Ignoring package %s as is not a dependency of the '
                'package being analyzed',
                package,
            )
            return

        self.user_mappings[package] = set(packages_provided)

        for single_package in packages_provided:
            self.reverse_user_mappings[single_package] = package

    def add_ignored_packages(self, packages):
        self.ignored_packages = {DottedName(package) for package in packages}

    def _all_requirements(self):
        all_requirements = self._requirements.copy()
        for extra in self._extras_requirements:
            all_requirements.update(self._extras_requirements[extra])

        return all_requirements

    def get_missing_imports(self):
        filters = (
            self._filter_out_testing_imports,
            self._filter_out_requirements,
            self._filter_out_ignored_imports,
        )
        missing = self._apply_filters(self.imports_used, filters)
        unique_imports = self._get_unique_imports(imports_list=missing)
        result = self._filter_user_mappings(unique_imports)
        return result

    def get_missing_test_imports(self):
        filters = (
            self._filter_out_only_testing_imports,
            self._filter_out_requirements,
            self._filter_out_test_requirements,
            self._filter_out_ignored_imports,
        )
        missing = self._apply_filters(self.imports_used, filters)
        unique_imports = self._get_unique_imports(imports_list=missing)
        result = self._filter_user_mappings(unique_imports)
        return result

    def get_unneeded_requirements(self):
        all_but_test_requirements = self._requirements.copy()
        for extra in self._extras_requirements:
            all_but_test_requirements.update(self._extras_requirements[extra])
        test_requirements = self._get_test_extra()
        for dotted_name in test_requirements:
            all_but_test_requirements.remove(dotted_name)
        filters = (
            self._filter_out_known_packages,
            self._filter_out_python_standard_library,
            self._filter_out_used_imports,
            self._filter_out_ignored_imports,
            self._filter_out_mappings,
        )
        unneeded = self._apply_filters(all_but_test_requirements, filters)
        unique_imports = self._get_unique_imports(imports_list=unneeded)
        return unique_imports

    def requirements_that_should_be_test_requirements(self):
        non_testing_filters = (self._filter_out_testing_imports,)
        non_testing_imports = self._apply_filters(
            self.imports_used,
            non_testing_filters,
        )
        complete_non_testing_imports = []
        for non_test_import in non_testing_imports:
            if non_test_import in self.reverse_user_mappings:
                meta_package = self.reverse_user_mappings[non_test_import]
                complete_non_testing_imports.append(meta_package)
                continue
            complete_non_testing_imports.append(non_test_import)

        requirements_not_used = [
            requirement
            for requirement in self._requirements
            if self._discard_if_found_obj_in_list(
                requirement,
                complete_non_testing_imports,
            )
        ]
        testing_filters = (
            self._filter_out_only_testing_imports,
            self._filter_out_own_package,
        )
        testing_imports = self._apply_filters(
            self.imports_used,
            testing_filters,
        )
        complete_test_imports = []
        for test_import in testing_imports:
            if test_import in self.reverse_user_mappings:
                meta_package = self.reverse_user_mappings[test_import]
                complete_test_imports.append(meta_package)
                continue
            complete_test_imports.append(test_import)
        should_be_test_requirements = [
            requirement
            for requirement in requirements_not_used
            if not self._discard_if_found_obj_in_list(
                requirement,
                complete_test_imports,
            )
        ]
        filters = (self._filter_out_ignored_imports,)
        skip_ignored = self._apply_filters(
            should_be_test_requirements,
            filters,
        )
        unique = self._get_unique_imports(imports_list=skip_ignored)
        return unique

    def get_unneeded_test_requirements(self):
        test_requirements = self._get_test_extra()
        if not test_requirements:
            return []

        filters = (
            self._filter_out_known_packages,
            self._filter_out_python_standard_library,
            self._filter_out_used_imports,
            self._filter_out_ignored_imports,
            self._filter_out_mappings_on_test,
        )
        unneeded = self._apply_filters(test_requirements, filters)
        unique_imports = self._get_unique_imports(imports_list=unneeded)
        return unique_imports

    @staticmethod
    def _apply_filters(objects, filters):
        """Filter a list of given objects through a list of given filters

        This helps encapsulating the filtering logic here while keeping the
        methods free from this clutter.

        Callees of this method need to provide a list of objects
        (usually a list of DottedNames) that will be processed by the provided
        list of filters.

        A filter, in this context, is a function that takes a DottedName as a
        single argument and returns a boolean value.

        If a given DottedName returns True in all filters,
        it will make it to the final result an be returned back.
        """
        result = []
        for single_object in objects:
            for filter_func in filters:
                if not filter_func(single_object):
                    break
            else:
                result.append(single_object)

        return result

    def _get_unique_imports(self, imports_list=None):
        if imports_list is None:
            imports_list = self.imports_used
        unique_dotted_names = list(set(imports_list))
        sorted_unique_dotted_names = sorted(unique_dotted_names)
        return sorted_unique_dotted_names

    def _filter_out_used_imports(self, dotted_name):
        return self._discard_if_found_obj_in_list(
            dotted_name,
            self.imports_used,
        )

    def _filter_out_mappings(self, meta_package):
        """Filter meta packages

        If it is not a meta package, let it continue the filtering.

        If it is a meta package, check if any of the packages it provides is
        used. If any of them it is, then filter it.
        """
        if meta_package not in self.user_mappings.keys():
            return True

        for dotted_name in self.user_mappings[meta_package]:
            if not self._discard_if_found_obj_in_list(dotted_name, self.imports_used):
                return False

        return True

    def _filter_out_mappings_on_test(self, meta_package):
        """Filter meta packages

        If it is not a meta package, let it continue the filtering.

        If it is a meta package, check if any of the packages it provides is
        used. If any of them it is, then filter it.
        """
        if meta_package not in self.user_mappings.keys():
            return True

        filters = (self._filter_out_only_testing_imports,)
        test_only_imports = self._apply_filters(self.imports_used, filters)

        for dotted_name in self.user_mappings[meta_package]:
            if not self._discard_if_found_obj_in_list(dotted_name, test_only_imports):
                return False

        return True

    def _filter_out_ignored_imports(self, dotted_name):
        return self._discard_if_found_obj_in_list(
            dotted_name,
            self.ignored_packages,
        )

    def _filter_out_known_packages(self, dotted_name):
        to_ignore = (
            DottedName('setuptools'),
            DottedName('pkg_resources'),
            DottedName('distribute'),
        )
        return self._discard_if_found_obj_in_list(dotted_name, to_ignore)

    @staticmethod
    def _filter_out_testing_imports(dotted_name):
        return not dotted_name.is_test

    @staticmethod
    def _filter_out_only_testing_imports(dotted_name):
        return dotted_name.is_test

    def _filter_out_own_package(self, dotted_name):
        return dotted_name not in self.own_dotted_name

    def _filter_out_requirements(self, dotted_name):
        return self._discard_if_found_obj_in_list(
            dotted_name,
            self._requirements,
        )

    def _filter_out_test_requirements(self, dotted_name):
        test_requirements = self._get_test_extra()
        if not test_requirements:
            return True

        return self._discard_if_found_obj_in_list(
            dotted_name,
            test_requirements,
        )

    def _filter_out_python_standard_library(self, dotted_name):
        std_library = self._build_std_library()
        return self._discard_if_found_obj_in_list(dotted_name, std_library)

    @staticmethod
    def _discard_if_found_obj_in_list(obj, obj_list):
        """Check if obj is on obj_list and return the opposite

        We are interested in filtering out requirements,
        so if a given object is on a given list,
        that object can be discarded as it's already in the list.
        """
        # this is a generator expression, nothing is computed until it is being
        # called some lines below
        iterable = (item for item in obj_list if obj in item)

        # any builtin stops consuming the iterator as soon as an element of the
        # iterable is True
        if any(iterable):
            # if 'obj' is found, the pipeline needs to stop and discard the
            # requirement
            return False

        # 'obj' was not found on the list, means that is fulfilling the
        # pipeline filters so far
        return True

    @staticmethod
    def _build_std_library():
        if PY_10_OR_HIGHER:
            # see https://github.com/jackmaney/python-stdlib-list/issues/55
            libraries = list(
                set(list(sys.stdlib_module_names) + list(sys.builtin_module_names))
            )
        else:
            from stdlib_list import stdlib_list

            python_version = sys.version_info
            libraries = stdlib_list(
                '{}.{}'.format(
                    python_version[0],
                    python_version[1],
                )
            )

        fake_std_libraries = [DottedName(x) for x in libraries]
        return fake_std_libraries

    def _get_test_extra(self):
        candidates = ('test', 'tests')
        for candidate in candidates:
            if candidate in self._extras_requirements:
                return self._extras_requirements[candidate]

        return []

    def _filter_user_mappings(self, dotted_names):
        """Remove dotted names that are in user mappings"""
        result = []
        for single_import in dotted_names:
            for provided_package in self.reverse_user_mappings:
                if single_import in provided_package:
                    logger.debug(
                        'Skip %s as is part of user mapping %s',
                        single_import,
                        self.reverse_user_mappings[provided_package],
                    )
                    break
            else:
                result.append(single_import)

        return result



import glob
import logging
import os
import sys

import pkg_resources
import toml
from cached_property import cached_property

from z3c.dependencychecker.db import ImportsDatabase
from z3c.dependencychecker.dotted_name import DottedName
from z3c.dependencychecker.modules import MODULES
from z3c.dependencychecker.utils import change_dir

logger = logging.getLogger(__name__)


class PackageMetadata:
    """Information related to a python source distribution

    It relies heavily on setuptools and pkg_resources APIs.
    """

    def __init__(self, path):
        logger.debug('Reading package metadata for %s...', path)
        self._path = path
        self._working_set = self._generate_working_set_with_ourselves()

    @cached_property
    def distribution_root(self):
        if 'setup.py' not in os.listdir(self._path):
            logger.error(
                'setup.py was not found in %s. '
                'Without it dependencychecker can not work.',
                self._path,
            )
            sys.exit(1)
        return self._path

    @cached_property
    def setup_py_path(self):
        return os.path.join(
            self.distribution_root,
            'setup.py',
        )

    @cached_property
    def package_dir(self):
        """Check where the .egg-info is located"""
        try_paths = [self.distribution_root]
        top_level_elements = [
            os.path.join(self.distribution_root, folder)
            for folder in os.listdir(self.distribution_root)
        ]
        try_paths += [folder for folder in top_level_elements if os.path.isdir(folder)]
        for path in try_paths:
            folder_found = self._find_egg_info_in_folder(path)
            if folder_found:
                logger.debug(
                    '.egg-info folder found at %s folder',
                    path,
                )
                return path

        logger.error('.egg-info folder could not be found')
        sys.exit(1)

    @staticmethod
    def _find_egg_info_in_folder(path):
        with change_dir(path):
            results = glob.glob('*.egg-info')
        return results

    @cached_property
    def egg_info_dir(self):
        results = self._find_egg_info_in_folder(self.package_dir)
        return os.path.join(
            self.package_dir,
            results[0],
        )

    @cached_property
    def name(self):
        path, filename = os.path.split(self.egg_info_dir)
        name = filename[: -len('.egg-info')]
        logger.debug(
            'Package name is %s',
            name,
        )

        return name

    def _generate_working_set_with_ourselves(self):
        """Use pkg_resources API to enable the package being analyzed

        This, enabling the package in pkg_resources, allows to extract
        information from .egg-info folders, for example, its requirements.
        """
        working_set = pkg_resources.WorkingSet()
        working_set.add_entry(self.package_dir)
        return working_set

    def _get_ourselves_from_working_set(self):
        """Use pkg_resources API to get a Requirement (in pkg_resources
        parlance) of the package being analyzed
        """
        ourselves = pkg_resources.Requirement.parse(self.name)
        try:
            package = self._working_set.find(ourselves)
            return package
        except ValueError:
            logger.error(
                'Package %s could not be found.\n'
                'You might need to put it in development mode,\n'
                'i.e. python setup.py develop',
                self.name,
            )
            sys.exit(1)

    def get_required_dependencies(self):
        this_package = self._get_ourselves_from_working_set()
        requirements = this_package.requires()
        for requirement in requirements:
            yield DottedName.from_requirement(
                requirement,
                file_path=self.setup_py_path,
            )

    def get_extras_dependencies(self):
        """Get this packages' extras dependencies defined in setup.py"""
        this_package = self._get_ourselves_from_working_set()

        for extra_name in this_package.extras:
            extra_requirements = this_package.requires(extras=(extra_name,))
            dotted_names = (
                DottedName.from_requirement(req, file_path=self.setup_py_path)
                for req in extra_requirements
            )
            yield extra_name, dotted_names

    @cached_property
    def top_level(self):
        path = os.path.join(
            self.egg_info_dir,
            'top_level.txt',
        )
        if not os.path.exists(path):
            logger.error(
                'top_level.txt could not be found on %s.\n'
                'It is needed for dependencychecker to work properly.',
                self.egg_info_dir,
            )
            sys.exit(1)

        with open(path) as top_level_file:
            content = top_level_file.read().strip()

        top_levels = []
        for candidate in content.split('\n'):
            possible_top_level = os.path.join(
                self.package_dir,
                candidate,
            )

            if os.path.exists(possible_top_level):
                logger.debug('Found top level %s', possible_top_level)
                top_levels.append(possible_top_level)
                continue

            single_module = f'{possible_top_level}.py'
            if os.path.exists(single_module):
                logger.debug('Found top level %s', single_module)
                top_levels.append(single_module)
                continue

            logger.warning(
                'Top level %s not found but referenced by top_level.txt',
                possible_top_level,
            )

        if top_levels:
            return top_levels

        logger.error(
            'There are paths found in %s%stop_level.txt that do not exist.\n'
            'Maybe you need to put the package in development again?',
            self.egg_info_dir,
            os.sep,
        )
        sys.exit(1)


class Package:
    """The python package that is being analyzed

    This class itself does not much per se, but connects the PackageMetadata
    with the ImportsDatabase, where the important bits are.
    """

    def __init__(self, path):
        self.metadata = PackageMetadata(path)
        self.imports = ImportsDatabase()
        self.imports.own_dotted_name = DottedName(self.metadata.name)

    def inspect(self):
        self.set_declared_dependencies()
        self.set_declared_extras_dependencies()
        self.set_user_mappings()
        self.analyze_package()

    def set_declared_dependencies(self):
        """Add this packages' dependencies defined in setup.py to the database"""
        self.imports.add_requirements(self.metadata.get_required_dependencies())

    def set_declared_extras_dependencies(self):
        """Add this packages' extras dependencies defined in setup.py to the
        database
        """
        for extra, dotted_names in self.metadata.get_extras_dependencies():
            self.imports.add_extra_requirements(extra, dotted_names)

    def set_user_mappings(self):
        """Add any user defined mapping

        They need to be on a pyproject.toml file within the table
        tool.dependencychecker.

        See tests/test_user_mappings.py for examples.
        """
        config = self._load_user_config()
        for package, packages_provided in config.items():
            if package == 'ignore-packages':
                if isinstance(packages_provided, list):
                    self.imports.add_ignored_packages(packages_provided)
                else:
                    logger.warning(
                        'ignore-packages key in pyproject.toml needs to '
                        'be a list, even for a single package to be ignored.'
                    )
            elif isinstance(packages_provided, list):
                self.imports.add_user_mapping(package, packages_provided)

    def analyze_package(self):
        for top_folder in self.metadata.top_level:
            logger.debug('Analyzing package top_level %s...', top_folder)
            for module_obj in MODULES:
                logger.debug(
                    'Starting analyzing files using %s...',
                    module_obj,
                )
                for source_file in module_obj.create_from_files(top_folder):
                    logger.debug(
                        'Searching dependencies (with %s) in file %s...',
                        module_obj.__name__,
                        source_file.path,
                    )
                    self.imports.add_imports(source_file.scan())

    def _load_user_config(self):
        config_file_path = os.sep.join(
            [self.metadata.distribution_root, 'pyproject.toml']
        )
        try:
            config = toml.load(config_file_path)
            return config['tool']['dependencychecker']
        except (OSError, KeyError):
            return {}



import os


def write_source_file_at(
    path_parts,
    filename='test.py',
    source_code='import unknown.dependency',
):
    folder_path = os.path.join(*path_parts)

    try:
        os.makedirs(folder_path)
    except OSError:
        pass

    file_path = os.path.join(folder_path, filename)
    with open(file_path, 'w') as new_file:
        new_file.write(source_code)

    return file_path


def get_requirements_names(database):
    return [requirement.name for requirement in database._requirements]


def get_extras_requirements_names(database):
    return database._extras_requirements.keys()


def get_requirements_names_for_extra(database, extra=''):
    return [requirement.name for requirement in database._extras_requirements[extra]]


def get_sorted_imports_paths(database):
    return sorted([x.file_path for x in database.imports_used])



import os

from z3c.dependencychecker.modules import ZCMLFile
from z3c.dependencychecker.tests.utils import write_source_file_at

ZCML_TEMPLATE = """
<configure
  xmlns="http://namespaces.zope.org/zope"
  xmlns:browser="http://namespaces.zope.org/browser"
  xmlns:genericsetup="http://namespaces.zope.org/genericsetup">

{0}

</configure>
"""

INCLUDE = '<include package="my.package.include" />'
ADAPTER_FACTORY = '<adapter factory="my.package.factory" />'
ADAPTER_FOR = '<adapter for="my.package.for" />'
ADAPTER_FOR_MULTIPLE = (
    '<adapter for="my.package.for another.package.foo yet.another.one" />'  # noqa
)
ADAPTER_PROVIDES = '<adapter provides="my.package.provides" />'
UTILITY_COMPONENT = '<utility component="my.package.component" />'
UTILITY_PROVIDES = '<utility provides="my.package.provides" />'
BROWSER_PAGE_CLASS = '<browser:page class="my.package.class" />'
BROWSER_PAGE_FOR = '<browser:page for="my.package.for" />'
BROWSER_PAGE_LAYER = '<browser:page layer="my.package.layer" />'
SUBSCRIBER_FOR = '<subscriber for="my.package.for" />'
SUBSCRIBER_FOR_MULTIPLE = """<subscriber for="my.package.for
                 another.package" />
"""
SUBSCRIBER_HANDLER = '<subscriber handler="my.package.handler" />'
SECURITYPOLICY_COMPONENT = '<securityPolicy component="my.package.component" />'  # noqa
GS_REGISTERPROFILE_PROVIDES = (
    '<genericsetup:registerProfile provides="my.package.provides" />'  # noqa
)
IGNORE_LOCAL_IMPORT = '<adapter factory=".package.factory" />'
ASTERISK = '<adapter for="*" />'


def _get_zcml_imports_on_file(folder, source):
    full_content = ZCML_TEMPLATE.format(source)
    temporal_file = write_source_file_at(
        (folder.strpath,),
        source_code=full_content,
        filename='configure.zcml',
    )

    zcml_file = ZCMLFile(folder.strpath, temporal_file)
    dotted_names = [x.name for x in zcml_file.scan()]
    return dotted_names


def test_create_from_files_nothing(minimal_structure):
    path, package_name = minimal_structure
    modules_found = list(ZCMLFile.create_from_files(path))
    assert len(modules_found) == 0


def test_create_from_files_deep_nested(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    write_source_file_at([src_path, 'a', 'b'], filename='configure.zcml')

    modules_found = list(ZCMLFile.create_from_files(src_path))
    assert len(modules_found) == 1


def test_include(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, INCLUDE)
    assert len(dotted_names) == 1


def test_include_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, INCLUDE)
    assert dotted_names == ['my.package.include']


def test_adapter_factory(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, ADAPTER_FACTORY)
    assert len(dotted_names) == 1


def test_adapter_factory_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, ADAPTER_FACTORY)
    assert dotted_names == ['my.package.factory']


def test_adapter_for(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, ADAPTER_FOR)
    assert len(dotted_names) == 1


def test_adapter_for_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, ADAPTER_FOR)
    assert dotted_names == ['my.package.for']


def test_adapter_for_multiple(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, ADAPTER_FOR_MULTIPLE)
    assert len(dotted_names) == 3


def test_adapter_for_multiple_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, ADAPTER_FOR_MULTIPLE)
    assert 'my.package.for' in dotted_names
    assert 'another.package.foo' in dotted_names
    assert 'yet.another.one' in dotted_names


def test_adapter_provides(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, ADAPTER_PROVIDES)
    assert len(dotted_names) == 1


def test_adapter_provides_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, ADAPTER_PROVIDES)
    assert dotted_names == ['my.package.provides']


def test_utility_component(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, UTILITY_COMPONENT)
    assert len(dotted_names) == 1


def test_utility_component_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, UTILITY_COMPONENT)
    assert dotted_names == ['my.package.component']


def test_utility_provides(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, UTILITY_PROVIDES)
    assert len(dotted_names) == 1


def test_utility_provides_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, UTILITY_PROVIDES)
    assert dotted_names == ['my.package.provides']


def test_browser_page_class(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, BROWSER_PAGE_CLASS)
    assert len(dotted_names) == 1


def test_browser_page_class_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, BROWSER_PAGE_CLASS)
    assert dotted_names == ['my.package.class']


def test_browser_page_for(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, BROWSER_PAGE_FOR)
    assert len(dotted_names) == 1


def test_browser_page_for_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, BROWSER_PAGE_FOR)
    assert dotted_names == ['my.package.for']


def test_browser_page_layer(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, BROWSER_PAGE_LAYER)
    assert len(dotted_names) == 1


def test_browser_page_layer_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, BROWSER_PAGE_LAYER)
    assert dotted_names == ['my.package.layer']


def test_subscriber_for(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, SUBSCRIBER_FOR)
    assert len(dotted_names) == 1


def test_subscriber_for_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, SUBSCRIBER_FOR)
    assert dotted_names == ['my.package.for']


def test_subscriber_for_multiple(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, SUBSCRIBER_FOR_MULTIPLE)
    assert len(dotted_names) == 2


def test_subscriber_for_multiple_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, SUBSCRIBER_FOR_MULTIPLE)
    assert dotted_names == [
        'my.package.for',
        'another.package',
    ]


def test_subscriber_handler(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, SUBSCRIBER_HANDLER)
    assert len(dotted_names) == 1


def test_subscriber_handler_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, SUBSCRIBER_HANDLER)
    assert dotted_names == ['my.package.handler']


def test_securitypolicy_component(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, SECURITYPOLICY_COMPONENT)
    assert len(dotted_names) == 1


def test_securitypolicy_component_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, SECURITYPOLICY_COMPONENT)
    assert dotted_names == ['my.package.component']


def test_gs_registerprofile_provides(tmpdir):
    dotted_names = _get_zcml_imports_on_file(
        tmpdir,
        GS_REGISTERPROFILE_PROVIDES,
    )
    assert len(dotted_names) == 1


def test_gs_registerprofile_provides_details(tmpdir):
    dotted_names = _get_zcml_imports_on_file(
        tmpdir,
        GS_REGISTERPROFILE_PROVIDES,
    )
    assert dotted_names == ['my.package.provides']


def test_ignore_local_import(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, IGNORE_LOCAL_IMPORT)
    assert len(dotted_names) == 0


def test_ignore_asterisk_import(tmpdir):
    dotted_names = _get_zcml_imports_on_file(tmpdir, ASTERISK)
    assert len(dotted_names) == 0



from z3c.dependencychecker.dotted_name import DottedName


def test_fallback():
    test_import = object()
    requirement = DottedName('plone.app.dexterity')
    result = test_import in requirement
    assert not result


def test_same_name_no_namespaces():
    test_import = DottedName('Plone')
    requirement = DottedName('Plone')
    result = test_import in requirement
    assert result


def test_different_name_no_namespaces():
    test_import = DottedName('Plone')
    requirement = DottedName('Zope')
    result = test_import in requirement
    assert not result


def test_no_substring_match_beginning():
    test_import = DottedName('plo')
    requirement = DottedName('plone.app.imaging')
    result = test_import in requirement
    assert not result


def test_no_substring_match_beginning_reverse_check():
    test_import = DottedName('reinout.happy')
    requirement = DottedName('re')
    result = test_import in requirement
    assert not result


def test_no_substring_match_middle():
    test_import = DottedName('ima')
    requirement = DottedName('plone.app.imaging')
    result = test_import in requirement
    assert not result


def test_no_substring_match_end():
    test_import = DottedName('imaging')
    requirement = DottedName('plone.app.imaging')
    result = test_import in requirement
    assert not result


def test_subpackage():
    test_import = DottedName('plone.app.imaging.interfaces.IImage')
    requirement = DottedName('plone.app.imaging')
    result = test_import in requirement
    assert result


def test_same_number_of_namespaces_but_different():
    test_import = DottedName('plone.app.dexterity')
    requirement = DottedName('plone.app.imaging')
    result = test_import in requirement
    assert not result


def test_share_only_part_of_namespace():
    test_import = DottedName('plone.app.dexterity.interfaces')
    requirement = DottedName('plone.app.imaging')
    result = test_import in requirement
    assert not result


def test_both_sides_are_tested():
    test_import = DottedName('plone.app.dexterity')
    requirement = DottedName('plone.app.imaging.interfaces')
    result = test_import in requirement
    assert not result



import os

from z3c.dependencychecker.dotted_name import DottedName
from z3c.dependencychecker.package import Package

EMPTY_FILE = ''
IGNORE_OTHER_KEYS = """[servers]
one = "1.2.3.4"
"""
ONLY_TOOL_TABLE = """[tool]
two = "5.6.7.8"
"""
EMPTY_DEPENDENCY_TABLE_NESTED = """[tool]
    [dependencycheker]
"""
EMPTY_DEPENDENCY_TABLE_MERGED = """[tool.dependencychecker]"""
ONE_MAPPING = """[tool.dependencychecker]
Zope2 = ["Products.Five", ]
"""
MORE_MAPPINGS = """[tool.dependencychecker]
Zope2 = ["Products.Five", ]
Zope4 = ["Products.Counter", "Products.Max" ]
"""
SUBTABLES = """[tool.dependencychecker]
Zope2 = ["Products.Five", ]
[tool.dependencychecker.options]
"""
IGNORE_PACKAGES_NO_LIST = """[tool.dependencychecker]
ignore-packages = 'oops'
"""
IGNORE_PACKAGES = """[tool.dependencychecker]
ignore-packages = ['django-toolbar', 'plone.reload']
"""


def _write_user_config(path, content):
    file_path = os.sep.join(
        [
            path,
            'pyproject.toml',
        ]
    )
    with open(file_path, 'w') as config_file:
        config_file.write(content)


def _update_requires_txt(path, package_name, packages):
    file_path = os.sep.join([path, f'{package_name}.egg-info', 'requires.txt'])
    with open(file_path, 'w') as config_file:
        for package in packages:
            config_file.write(f'{package}\n')


def test_no_file(minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    package.set_user_mappings()
    assert package.imports.user_mappings == {}


def test_empty_file(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, EMPTY_FILE)
    package = Package(path)
    package.set_user_mappings()
    assert package.imports.user_mappings == {}


def test_ignore_other_keys(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, IGNORE_OTHER_KEYS)
    package = Package(path)
    package.set_user_mappings()
    assert package.imports.user_mappings == {}


def test_only_tool_table(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, ONLY_TOOL_TABLE)
    package = Package(path)
    package.set_user_mappings()
    assert package.imports.user_mappings == {}


def test_empty_dependency_table_nested(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, EMPTY_DEPENDENCY_TABLE_NESTED)
    package = Package(path)
    package.set_user_mappings()
    assert package.imports.user_mappings == {}


def test_empty_dependency_table_merged(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, EMPTY_DEPENDENCY_TABLE_MERGED)
    package = Package(path)
    package.set_user_mappings()
    assert package.imports.user_mappings == {}


def test_ignore_user_mappings(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, ONE_MAPPING)
    package = Package(path)
    package.inspect()

    assert len(package.imports.user_mappings) == 0


def test_one_user_mapping(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, ONE_MAPPING)
    _update_requires_txt(
        path,
        package_name,
        [
            'Zope2',
        ],
    )
    package = Package(path)
    package.inspect()
    five_dotted_name = DottedName('Products.Five')
    zope_dotted_name = DottedName('Zope2')

    assert len(package.imports.user_mappings) == 1
    assert zope_dotted_name in package.imports.user_mappings

    mappings = package.imports.user_mappings[zope_dotted_name]
    assert len(mappings) == 1
    assert {
        five_dotted_name,
    } == mappings


def test_filter_out_mappings_on_test(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, ONE_MAPPING)
    _update_requires_txt(
        path,
        package_name,
        ['plone.reload', '[test]', 'Zope2'],
    )
    package = Package(path)
    package.inspect()
    five_dotted_name = DottedName('Products.Five')
    plone_reload_dotted_name = DottedName('plone.reload')
    zope_dotted_name = DottedName('Zope2')

    assert len(package.imports.user_mappings) == 1
    assert zope_dotted_name in package.imports.user_mappings

    # The next two aren'd in the mapping, so they shouldn't be filtered out:
    assert package.imports._filter_out_mappings_on_test(plone_reload_dotted_name)
    assert package.imports._filter_out_mappings_on_test(five_dotted_name)
    # This one is in the mapping, but shouldn't be filtered out as it is a test-only import.
    assert package.imports._filter_out_mappings_on_test(zope_dotted_name)


def test_more_user_mappings(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, MORE_MAPPINGS)
    _update_requires_txt(
        path,
        package_name,
        [
            'Zope2',
            'Zope4',
        ],
    )
    package = Package(path)
    package.inspect()
    zope2_dotted_name = DottedName('Zope2')
    zope4_dotted_name = DottedName('Zope4')

    assert len(package.imports.user_mappings) == 2
    assert zope2_dotted_name in package.imports.user_mappings
    assert zope4_dotted_name in package.imports.user_mappings

    assert len(package.imports.user_mappings[zope4_dotted_name]) == 2


def test_subtables(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, SUBTABLES)
    _update_requires_txt(
        path,
        package_name,
        [
            'Zope2',
        ],
    )
    package = Package(path)
    package.inspect()
    zope2_dotted_name = DottedName('Zope2')

    assert len(package.imports.user_mappings) == 1
    assert zope2_dotted_name in package.imports.user_mappings


def test_ignore_packages_no_list(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, IGNORE_PACKAGES_NO_LIST)
    package = Package(path)
    package.inspect()
    assert len(package.imports.user_mappings) == 0


def test_ignore_packages(minimal_structure):
    path, package_name = minimal_structure
    _write_user_config(path, IGNORE_PACKAGES)
    _update_requires_txt(
        path,
        package_name,
        ['Zope2', 'plone.reload', 'django-toolbar'],
    )
    package = Package(path)
    package.inspect()
    django_toolbar = DottedName('django-toolbar')
    plone_reload = DottedName('plone.reload')
    ignored_packages = package.imports.ignored_packages

    assert len(package.imports.user_mappings) == 0
    assert len(ignored_packages) == 2
    assert django_toolbar in ignored_packages
    assert plone_reload in ignored_packages



from unittest import mock

from z3c.dependencychecker.package import Package
from z3c.dependencychecker.report import Report
from z3c.dependencychecker.tests.utils import write_source_file_at


def test_print_header_nothing(capsys):
    fake_package = mock.MagicMock()
    report = Report(fake_package)
    report._print_header(None)
    out, err = capsys.readouterr()
    assert out == ''


def test_print_header_text(capsys):
    fake_project = mock.MagicMock()
    report = Report(fake_project)
    report._print_header('hi')
    out, err = capsys.readouterr()

    assert 'hi' in out
    assert '==' in out
    assert '===' not in out


def test_print_report(capsys, minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    report = Report(package)
    report.print_report()
    out, err = capsys.readouterr()
    assert '======' not in out  # there are no errors on the dependencies
    assert '' == err


def test_missing_requirements_nothing(capsys):
    fake_project = mock.MagicMock()
    report = Report(fake_project)
    report.missing_requirements()
    out, err = capsys.readouterr()
    assert out == ''


def test_missing_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at((path, package_name), '__init__.py', 'import another.package')

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    assert 'Missing requirements\n====================' in out
    assert 'another.package' in out


def test_missing_requirements_with_user_mapping(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import Products.Five.browser.views.BrowserView',
    )
    write_source_file_at(
        (path, f'{package_name}.egg-info'),
        'requires.txt',
        'Zope2',
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(
            [
                '[tool.dependencychecker]',
                'Zope2 = ["Four", "Products.Five", "Three" ]',
            ]
        ),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    assert 'Missing requirements\n====================' not in out
    assert 'Products.Five' not in out
    assert 'Zope2' not in out


def test_missing_requirements_with_ignored_packages(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import Products.Five.browser.views.BrowserView',
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(
            [
                '[tool.dependencychecker]',
                'ignore-packages = ["Four", "Products.Five", "Three" ]',
            ]
        ),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_requirements()
    out, err = capsys.readouterr()

    assert 'Missing requirements\n====================' not in out
    assert 'Products.Five' not in out
    assert 'Four' not in out
    assert 'Three' not in out


def test_missing_test_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        '',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'import another.package',
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert 'Missing test requirements\n=========================' in out
    assert 'another.package' in out


def test_missing_test_requirements_with_user_mapping(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        '',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'import Products.Five.browser.views.BrowserView',
    )
    write_source_file_at(
        (path, f'{package_name}.egg-info'),
        'requires.txt',
        'Zope2',
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(['[tool.dependencychecker]', 'Zope2 = ["Products.Five" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert '' == out
    assert 'Missing requirements\n====================' not in out
    assert 'Products.Five' not in out
    assert 'Zope2' not in out


def test_missing_test_requirements_with_user_mapping_on_test_extra(
    capsys, minimal_structure
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        '',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'import Products.Five.browser.views.BrowserView',
    )
    write_source_file_at(
        (path, f'{package_name}.egg-info'),
        'requires.txt',
        '\n'.join(['[test]', 'Zope2']),
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(['[tool.dependencychecker]', 'Zope2 = ["Products.Five" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert '' == out
    assert 'Missing requirements\n====================' not in out
    assert 'Products.Five' not in out
    assert 'Zope2' not in out


def test_missing_test_requirements_with_ignored_packages(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        '',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'import Products.Five.browser.views.BrowserView',
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(
            [
                '[tool.dependencychecker]',
                'ignore-packages = ["Products.Five" ]',
            ]
        ),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.missing_test_requirements()
    out, err = capsys.readouterr()

    assert '' == out
    assert 'Missing requirements\n====================' not in out
    assert 'Products.Five' not in out


def test_unneeded_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import this.package',
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded requirements\n=====================' in out
    assert 'one' in out
    assert 'two' in out


def test_unneeded_requirements_with_ignored_packages(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import this.package',
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(
            [
                '[tool.dependencychecker]',
                'ignore-packages = ["one" ]',
            ]
        ),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded requirements\n=====================' in out
    assert 'one' not in out
    assert 'two' in out


def test_unneeded_requirements_with_user_mapping(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'from BTrees import Tree',
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(['[tool.dependencychecker]', '"ZODB3" = ["BTrees" ]']),
    )
    write_source_file_at(
        (path, f'{package_name}.egg-info'),
        'requires.txt',
        '\n'.join(['ZODB3', 'setuptools', 'one']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded requirements\n=====================' in out
    assert 'ZODB3' not in out
    assert 'one' in out


def test_unneeded_requirements_with_user_mapping2(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'from Plone import Tree',
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(['[tool.dependencychecker]', '"ZODB3" = ["BTrees" ]']),
    )
    write_source_file_at(
        (path, f'{package_name}.egg-info'),
        'requires.txt',
        '\n'.join(['ZODB3', 'setuptools', 'one']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded requirements\n=====================' in out
    assert 'ZODB3' in out
    assert 'one' in out


def test_unneeded_test_requirements(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join(['[test]', 'pytest', 'mock']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded test requirements\n==========================' in out
    assert 'pytest' in out
    assert 'mock' in out


def test_unneeded_test_requirements_with_ignore_packages(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join(['[test]', 'pytest', 'mock']),
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(['[tool.dependencychecker]', 'ignore-packages = ["mock" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded test requirements\n==========================' in out
    assert 'pytest' in out
    assert 'mock' not in out


def test_unneeded_test_requirements_with_user_mappings(capsys, minimal_structure):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join(['[test]', 'pytest', 'ZODB3']),
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(['[tool.dependencychecker]', '"ZODB3" = ["BTrees" ]']),
    )
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        '',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'from BTrees import Tree',
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert 'Unneeded test requirements\n==========================' in out
    assert 'pytest' in out
    assert 'ZODB3' not in out
    assert 'BTrees' not in out


def test_unneeded_test_requirements_no_tests_requirements(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.unneeded_test_requirements()
    out, err = capsys.readouterr()

    assert '' == out


def test_requirements_that_should_be_test_requirements(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import one',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        'import two',
    )
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join(['one', 'two', '[test]', 'pytest', 'mock']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.requirements_that_should_be_test_requirements()
    out, err = capsys.readouterr()

    assert (
        'Requirements that should be test requirements\n'
        '=============================================' in out
    )
    assert 'two' in out


def test_requirements_that_should_be_test_requirements_with_ignored_packages(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import one',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        '\n'.join(['import two', 'import three']),
    )
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join(['one', 'two', 'three', '[test]', 'pytest', 'mock']),
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(['[tool.dependencychecker]', 'ignore-packages = ["two" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.requirements_that_should_be_test_requirements()
    out, err = capsys.readouterr()

    assert (
        'Requirements that should be test requirements\n'
        '=============================================' in out
    )
    assert 'three' in out
    assert 'two' not in out


def test_requirements_that_should_be_test_requirements_with_user_mappings(
    capsys,
    minimal_structure,
):
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        'import one',
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        '\n'.join(['import two', 'import BTrees']),
    )
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join(['one', 'two', 'ZODB3', '[test]', 'pytest', 'mock']),
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(['[tool.dependencychecker]', '"ZODB3" = ["BTrees" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.requirements_that_should_be_test_requirements()
    out, err = capsys.readouterr()

    assert (
        'Requirements that should be test requirements\n'
        '=============================================' in out
    )
    assert 'ZODB3' in out
    assert 'two' in out
    assert 'one' not in out


def test_requirements_that_should_be_test_requirements_with_user_mapping2(
    capsys,
    minimal_structure,
):
    """Test that user mappings used in both regular and test imports are
    not downgraded to test imports only.
    """
    path, package_name = minimal_structure
    write_source_file_at(
        (path, package_name),
        '__init__.py',
        '\n'.join(['import one', 'import two', 'import part.of.meta']),
    )
    write_source_file_at(
        (path, package_name, 'tests'),
        '__init__.py',
        '\n'.join(['import three', 'import four', 'import part.of.meta']),
    )
    write_source_file_at(
        (path, package_name + '.egg-info'),
        'requires.txt',
        '\n'.join(['one', 'two', 'three', 'meta-package', '[test]', 'four']),
    )
    write_source_file_at(
        (path,),
        'pyproject.toml',
        '\n'.join(['[tool.dependencychecker]', 'meta-package = ["part.of.meta" ]']),
    )

    package = Package(path)
    package.inspect()
    report = Report(package)
    report.requirements_that_should_be_test_requirements()
    out, err = capsys.readouterr()

    assert (
        'Requirements that should be test requirements\n'
        '=============================================' in out
    )
    assert 'three' in out
    assert 'one' not in out
    assert 'two' not in out
    assert 'part.of.meta' not in out
    assert 'meta-package' not in out


def test_print_notice(capsys, minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    report = Report(package)
    report.print_notice()
    out, err = capsys.readouterr()
    assert 'Note: ' in out
    assert '' == err


def test_exit_status_set(minimal_structure):
    path, package_name = minimal_structure
    package = Package(path)
    package.inspect()
    report = Report(package)
    assert report.exit_status == 0
    report.print_report()
    assert report.exit_status == 1



import hashlib
from functools import total_ordering

from cached_property import cached_property


@total_ordering
class DottedName:
    def __init__(
        self,
        name,
        file_path=None,
        is_test=False,
    ):
        self.name = name
        self.safe_name = name.lower().replace('-', '_')

        self.file_path = file_path
        self.is_test = is_test

    @classmethod
    def from_requirement(cls, requirement, file_path=None):
        """A requirement in this method's context is a
        pkg_resources.Requirement
        """
        return cls(
            requirement.project_name,
            file_path=file_path,
        )

    @cached_property
    def namespaces(self):
        return self.safe_name.split('.')

    @cached_property
    def is_namespaced(self):
        return bool(len(self.namespaces) - 1)

    def __lt__(self, other):
        if not isinstance(other, DottedName):
            return NotImplemented
        return self.name < other.name

    def __eq__(self, other):
        if not isinstance(other, DottedName):
            return NotImplemented

        return self.safe_name == other.safe_name

    def __repr__(self):
        return f'<DottedName {self.name}>'

    def __hash__(self):
        digest = hashlib.sha256(self.safe_name.encode()).hexdigest()
        return int(digest, 16) % 10**8

    def __contains__(self, item):
        """Check if self is in item or the other way around

        As we can never know which one is actually the requirement/pypi name
        and which one the import, check if one fits inside the other.

        So ``x in s`` and ``s in x`` should return the same in this
        implementation.
        """
        if not isinstance(item, DottedName):
            return False

        if self.safe_name == item.safe_name:
            return True

        # note that zip makes two different sized iterables have the same size
        # i.e. [1,2,3] and [4,5,6,7,8] will only loop through [1,4] [2,5]
        # and [3,6]. The other elements (7 and 8) will be totally ignored.
        # That's a nice trick to ensure that all the namespace is shared.
        for part1, part2 in zip(self.namespaces, item.namespaces):
            if part1 != part2:
                return False

        return True



import logging
import os
import sys
import tempfile
from unittest import mock

from z3c.dependencychecker.main import (
    _version,
    get_path,
    main,
    parse_command_line,
    set_log_level,
)
from z3c.dependencychecker.utils import change_dir


def test_usage_non_existing_option():
    arguments = ['dependencychecker', '--non-existing']
    sys_exit = False
    try:
        with mock.patch.object(sys, 'argv', arguments):
            parse_command_line()
    except SystemExit:
        sys_exit = True

    assert sys_exit


def test_usage_set_verbose():
    arguments = ['dependencychecker', '--verbose']
    with mock.patch.object(sys, 'argv', arguments):
        options, args = parse_command_line()

    assert options.verbose


def test_version():
    assert _version().endswith('.dev0')


def test_set_debug_logging():
    _cleanup_logging_handlers()
    set_log_level(True)
    assert logging.getLogger().getEffectiveLevel() == logging.DEBUG


def test_set_info_logging():
    _cleanup_logging_handlers()
    set_log_level(False)
    assert logging.getLogger().getEffectiveLevel() == logging.INFO


def _cleanup_logging_handlers():
    """The first iteration of this code was done as a pytest fixture,
    but introducing the pytest-catchlog plugin made it fail.

    On top of that, the loop inside a loop in this function is due to some
    sort of stickiness from pytest-catchlog handler that fails to be removed on
    a first iteration...
    """
    root_logger = logging.getLogger()
    while len(root_logger.handlers) > 0:
        handlers = root_logger.handlers
        for handler in handlers:
            root_logger.removeHandler(handler)


def test_get_path_no_path_given():
    """If no path is given as an argument the cwd is used"""
    folder = tempfile.mkdtemp()
    arguments = []
    with change_dir(folder):
        path = get_path(arguments)

    # realpath() below is used to work around a small mac tempdir naming
    # issue. /var/folders/xyz is the tempdir, but after cd'ing you're in
    # /private/var/folders/xyz...
    assert os.path.realpath(path) == os.path.realpath(folder)


def test_get_path_path_given():
    """If a path is given, that's the path being returned"""
    folder = tempfile.mkdtemp()
    arguments = [folder]
    path = get_path(arguments)

    assert path == folder


def test_get_path_given_path_not_a_folder():
    """If the path given is not a folder, get_path exits right away"""
    temporary_file = tempfile.NamedTemporaryFile()
    arguments = [temporary_file.name]
    sys_exit = False
    try:
        get_path(arguments)
    except SystemExit:
        sys_exit = True

    assert sys_exit


def test_exit_zero_not_set(minimal_structure):
    path, _ = minimal_structure

    arguments = ['dependencychecker', path]
    exit_code = 0
    try:
        with mock.patch.object(sys, 'argv', arguments):
            main()
    except SystemExit as exc:
        exit_code = exc.code

    assert exit_code == 1


def test_exit_zero_set(minimal_structure):
    path, _ = minimal_structure

    arguments = ['dependencychecker', '--exit-zero', path]
    exit_code = 0
    try:
        with mock.patch.object(sys, 'argv', arguments):
            main()
    except SystemExit as exc:
        exit_code = exc.code

    assert exit_code == 0



import codecs
import os.path

from setuptools import find_packages, setup

version = '2.8'


def read(filename):
    try:
        with codecs.open(filename, encoding='utf-8') as f:
            return unicode(f.read())
    except NameError:
        with open(filename, encoding='utf-8') as f:
            return f.read()


long_description = f"""
{read('README.rst')}

{read(os.path.join('z3c', 'dependencychecker', 'USAGE.rst'))}

{read('CHANGES.rst')}
"""


setup(
    name='z3c.dependencychecker',
    version=version,
    description='Reports on missing or unneeded setup.py dependencies',
    long_description=long_description,
    long_description_content_type='text/x-rst; charset=UTF-8',
    # Get strings from https://pypi.org/classifiers/
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Quality Assurance',
    ],
    keywords='dependencies requirements missing imports',
    author='Reinout van Rees',
    author_email='reinout@vanrees.org',
    url='https://github.com/reinout/z3c.dependencychecker',
    license='BSD',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['z3c'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'setuptools',
        'stdlib-list; python_version < "3.10"',
        'cached-property',
        'toml',
    ],
    extras_require={
        'test': ['pytest'],
    },
    entry_points={
        'console_scripts': [
            'dependencychecker = z3c.dependencychecker.main:main',
        ]
    },
)



import os

from z3c.dependencychecker.modules import DocFiles
from z3c.dependencychecker.tests.utils import write_source_file_at

NO_DOC = """
Random title
============

Random paragraph
"""
INVALID_PYTHON = """
Random title
============

Random paragraph
>>> and some text here that breaks the parser
"""
SINGLE_IMPORT = """
Random title
============

Random paragraph
>>> import zope.annotation
"""
MULTIPLE_IMPORTS_SAME_LINE = """
Random title
============

Random paragraph
>>> from zope.interface import Interface, Attribute
"""
MULTIPLE_IMPORTS_DIFFERENT_LINES = '''
class MyClass(object):',
    def test(self):',
        """Docstring with code to be evaluated.

        >>> from zope.component import adapter
        >>> from zope.component import utility
        """
'''
MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN = '''
class MyClass(object):',
    def test(self):',
        """Docstring with code to be evaluated.

        >>> from zope.component import adapter
        >>> this should  be skipped and next line processed happily
        >>> from zope.component import utility
        """'
'''


def _get_dependencies_on_file(folder, source):
    temporal_file = write_source_file_at(
        (folder.strpath,),
        source_code=source,
    )

    doc_file = DocFiles(folder.strpath, temporal_file)
    dotted_names = [x.name for x in doc_file.scan()]
    return dotted_names


def test_create_from_files_nothing(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    modules_found = list(DocFiles.create_from_files(src_path))
    assert len(modules_found) == 0


def test_create_from_files_deep_nested_txt(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    write_source_file_at(
        [src_path, 'a', 'b', 'c'],
        filename='bla.txt',
    )

    modules_found = list(DocFiles.create_from_files(src_path))
    assert len(modules_found) == 1


def test_create_from_files_deep_nested_rst(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    write_source_file_at(
        [src_path, 'a', 'b', 'c'],
        filename='bla.rst',
    )

    modules_found = list(DocFiles.create_from_files(src_path))
    assert len(modules_found) == 1


def test_no_code(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, NO_DOC)
    assert len(dotted_names) == 0


def test_invalid_code_ignored(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, INVALID_PYTHON)
    assert len(dotted_names) == 0


def test_code_found(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, SINGLE_IMPORT)
    assert len(dotted_names) == 1


def test_code_found_details(tmpdir):
    dotted_names = _get_dependencies_on_file(tmpdir, SINGLE_IMPORT)
    assert dotted_names == ['zope.annotation']


def test_always_testing(tmpdir):
    temporal_file = write_source_file_at(
        (tmpdir.strpath,), source_code='>>> import foo', filename='file.rst'
    )

    doc_file = DocFiles(tmpdir.strpath, temporal_file)
    dotted_names = list(doc_file.scan())
    assert len(dotted_names) == 1
    assert dotted_names[0].is_test


def test_multiple_imports_same_line(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_SAME_LINE,
    )
    assert len(dotted_names) == 2


def test_multiple_imports_same_line_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_SAME_LINE,
    )
    assert 'zope.interface.Interface' in dotted_names
    assert 'zope.interface.Attribute' in dotted_names


def test_multiple_imports_different_lines(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES,
    )
    assert len(dotted_names) == 2


def test_multiple_imports_different_lines_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES,
    )

    assert 'zope.component.adapter' in dotted_names
    assert 'zope.component.utility' in dotted_names


def test_multiple_imports_with_invalid_lines(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN,
    )
    assert len(dotted_names) == 2


def test_multiple_imports_with_invalid_lines_details(tmpdir):
    dotted_names = _get_dependencies_on_file(
        tmpdir,
        MULTIPLE_IMPORTS_DIFFERENT_LINES_WITH_INVALID_CODE_LINES_BETWEEN,
    )

    assert 'zope.component.adapter' in dotted_names
    assert 'zope.component.utility' in dotted_names



from z3c.dependencychecker.db import ImportsDatabase
from z3c.dependencychecker.dotted_name import DottedName
from z3c.dependencychecker.tests.utils import (
    get_extras_requirements_names,
    get_requirements_names,
    get_requirements_names_for_extra,
)


def test_no_dependencies():
    database = ImportsDatabase()
    assert isinstance(database._requirements, set)
    assert len(database._requirements) == 0


def test_add_dependencies():
    database = ImportsDatabase()
    database.add_requirements([DottedName('one'), DottedName('two')])

    names = get_requirements_names(database)
    assert len(names) == 2
    assert 'one' in names
    assert 'two' in names


def test_declared_dependencies_no_duplicates():
    """Check that if there are duplicated requirements,
    they are correctly filtered and only one remains
    """
    database = ImportsDatabase()
    database.add_requirements(
        [DottedName('one'), DottedName('two'), DottedName('three'), DottedName('one')]
    )

    requirements = get_requirements_names(database)
    assert len(requirements) == 3


def test_no_extras():
    database = ImportsDatabase()
    assert len(get_extras_requirements_names(database)) == 0


def test_extras_requirements_key_names():
    database = ImportsDatabase()
    database.add_extra_requirements('test', [DottedName('one')])
    database.add_extra_requirements('plone', [DottedName('two')])

    extras = get_extras_requirements_names(database)
    assert len(extras) == 2
    assert 'test' in extras
    assert 'plone' in extras


def test_add_extra_dependencies():
    database = ImportsDatabase()
    imports = [
        DottedName('one'),
        DottedName('two'),
    ]
    database.add_extra_requirements('test', imports)

    test_extras = get_requirements_names_for_extra(database, extra='test')
    assert len(test_extras) == 2
    assert 'one' in test_extras
    assert 'two' in test_extras


def test_warning_extra_declared_twice(caplog):
    """Show a warning if an extra is declared twice on setup.py"""
    database = ImportsDatabase()
    database.add_extra_requirements('test', [DottedName('one')])
    database.add_extra_requirements('test', [DottedName('two')])

    last_log = caplog.records[-1]
    message = 'extra requirement "test" is declared twice in setup.py'
    assert message == last_log.message


def test_direct_requirements_filtered_on_extras():
    """Check that direct requirements are filtered when added to an extra"""
    database = ImportsDatabase()
    database.add_requirements([DottedName('one'), DottedName('two')])
    database.add_extra_requirements('test', [DottedName('one'), DottedName('three')])

    test_extras = get_requirements_names_for_extra(database, extra='test')
    assert len(test_extras) == 1
    assert 'three' in test_extras
    assert 'one' not in test_extras


def test_filter_duplicate_extra_requirements():
    """If an extra is declared twice, duplicate requirements are filtered"""
    database = ImportsDatabase()
    database.add_extra_requirements('test', [DottedName('one'), DottedName('three')])
    database.add_extra_requirements('test', [DottedName('two'), DottedName('three')])

    test_extras = get_requirements_names_for_extra(database, extra='test')
    assert len(test_extras) == 3
    assert 'one' in test_extras
    assert 'two' in test_extras
    assert 'three' in test_extras


def test_no_used_imports():
    database = ImportsDatabase()
    assert len(database.imports_used) == 0


def test_add_used_imports():
    database = ImportsDatabase()
    database.own_dotted_name = DottedName('something-else')
    database.add_imports([DottedName('one'), DottedName('two')])
    assert len(database.imports_used) == 2


def test_unique_imports(minimal_database):
    minimal_database.add_imports(
        [
            DottedName('zope.component.ISite'),
            DottedName('zope.component'),
            DottedName('zope.component'),
        ]
    )

    dotted_names = minimal_database._get_unique_imports()
    names = [x.name for x in dotted_names]

    assert len(dotted_names) == 2
    assert 'zope.component' in names
    assert 'zope.component.ISite' in names


def test_unique_and_sorted_imports(minimal_database):
    minimal_database.add_imports(
        [DottedName('zope.c'), DottedName('zope.a'), DottedName('zope.b')]
    )

    dotted_names = minimal_database._get_unique_imports()

    assert len(dotted_names) == 3
    assert dotted_names[0].name == 'zope.a'
    assert dotted_names[1].name == 'zope.b'
    assert dotted_names[2].name == 'zope.c'


def test_filter_out_known_packages(minimal_database):
    dotted_name = DottedName('bla')
    result = minimal_database._filter_out_known_packages(dotted_name)
    assert result is True


def test_filter_out_known_packages_filtered(minimal_database):
    dotted_name = DottedName('setuptools')
    result = minimal_database._filter_out_known_packages(dotted_name)
    assert result is False


def test_filter_out_known_packages_filtered_nested(minimal_database):
    dotted_name = DottedName('setuptools.my.package')
    result = minimal_database._filter_out_known_packages(dotted_name)
    assert result is False


def test_filter_out_known_packages_filtered_nested_other(minimal_database):
    dotted_name = DottedName('pkg_resources.my.package')
    result = minimal_database._filter_out_known_packages(dotted_name)
    assert result is False


def test_filter_out_testing_imports(minimal_database):
    dotted_name = DottedName('bla', is_test=True)
    result = minimal_database._filter_out_testing_imports(dotted_name)
    assert result is False


def test_filter_out_testing_imports_no_testing_import(minimal_database):
    dotted_name = DottedName('bla', is_test=False)
    result = minimal_database._filter_out_testing_imports(dotted_name)
    assert result is True


def test_filter_out_own_package(minimal_database):
    minimal_database.own_dotted_name = DottedName('zope.component')
    result = minimal_database._filter_out_own_package(DottedName('bla'))
    assert result is True


def test_filter_out_own_package_filtered(minimal_database):
    minimal_database.own_dotted_name = DottedName('zope.component')
    result = minimal_database._filter_out_own_package(
        DottedName('zope.component'),
    )
    assert result is False


def test_filter_out_own_package_if_subpackage(minimal_database):
    minimal_database.own_dotted_name = DottedName('zope.component')
    subpackage = DottedName('zope.component.adapter')
    result = minimal_database._filter_out_own_package(subpackage)
    assert result is False


def test_filter_out_requirements(minimal_database):
    pkg = DottedName('zope.component')
    minimal_database.add_requirements([pkg])
    result = minimal_database._filter_out_requirements(pkg)
    assert result is False


def test_filter_out_std_library(minimal_database):
    pkg = DottedName('os.path.join')
    result = minimal_database._filter_out_python_standard_library(pkg)
    assert result is False


def test_filter_out_requirements_keep_other(minimal_database):
    pkg1 = DottedName('zope.component')
    pkg2 = DottedName('zope.interface')

    minimal_database.add_imports([pkg1, pkg2])
    minimal_database.add_requirements([pkg1])
    dotted_names = minimal_database.get_missing_imports()

    assert len(dotted_names) == 1
    assert pkg2 is dotted_names[0]


def test_filter_out_test_requirements_in_extra(minimal_database):
    dotted_name = DottedName('bli.blu.bla')
    minimal_database.add_extra_requirements(
        'test',
        (DottedName('bla'), DottedName('bli')),
    )
    result = minimal_database._filter_out_test_requirements(dotted_name)
    assert result is False


def test_filter_out_test_requirements_in_extra_variant(minimal_database):
    dotted_name = DottedName('bli.blu.bla')
    minimal_database.add_extra_requirements(
        'tests',
        (DottedName('bla'), DottedName('bli')),
    )
    result = minimal_database._filter_out_test_requirements(dotted_name)
    assert result is False


def test_get_imports_used_filter_subpackage(minimal_database):
    subpkg1 = DottedName('zope.component.adapter')
    subpkg2 = DottedName('zope.component.another.one')

    minimal_database.add_imports([subpkg1, subpkg2])
    minimal_database.add_requirements([DottedName('zope.component')])
    dotted_names = minimal_database.get_missing_imports()

    assert len(dotted_names) == 0


def test_get_imports_used_filter_std_library(minimal_database):
    minimal_database.add_imports(
        [
            DottedName('zope.component'),
            DottedName('os.path.join'),
            DottedName('sys.version_info'),
        ]
    )

    dotted_names = minimal_database.get_missing_imports()

    assert len(dotted_names) == 1
    assert dotted_names[0].name == 'zope.component'


def test_get_missing_testing_imports(minimal_database):
    test_import = DottedName('zope.component', is_test=True)
    regular_import = DottedName('zope.interface', is_test=False)
    minimal_database.add_imports([test_import, regular_import])

    dotted_names = minimal_database.get_missing_test_imports()

    assert len(dotted_names) == 1
    assert dotted_names[0] is test_import


def test_get_missing_testing_imports_filter_test_extras(minimal_database):
    test_import1 = DottedName('zope.component', is_test=True)
    test_import2 = DottedName('zope.interface', is_test=True)
    minimal_database.add_imports([test_import1, test_import2])
    minimal_database.add_extra_requirements('test', (test_import1,))

    dotted_names = minimal_database.get_missing_test_imports()

    assert len(dotted_names) == 1
    assert dotted_names[0] is test_import2


def test_get_test_extra(minimal_database):
    dotted_name = DottedName('bla')
    minimal_database.add_extra_requirements('test', (dotted_name,))
    result = minimal_database._get_test_extra()
    assert len(result) == 1
    assert dotted_name in result


def test_get_test_extra_plural(minimal_database):
    dotted_name = DottedName('bla')
    minimal_database.add_extra_requirements('tests', (dotted_name,))
    result = minimal_database._get_test_extra()
    assert len(result) == 1
    assert dotted_name in result


def test_get_test_extra_no_extra(minimal_database):
    dotted_name = DottedName('bla')
    minimal_database.add_extra_requirements('other', (dotted_name,))
    result = minimal_database._get_test_extra()
    assert result == []



import os
import tempfile

from z3c.dependencychecker.modules import DjangoSettings
from z3c.dependencychecker.tests.utils import write_source_file_at

RANDOM_CODE = 'from zope.component import adapter'
RANDOM_ASSIGNMENT = 'a = 3'
APPS_ASSIGNMENT_TO_STRING = 'INSTALLED_APPS = "random"'
APPS_ASSIGNMENT_TO_TUPLE = 'INSTALLED_APPS = ("random1", "random2", )'
APPS_ASSIGNMENT_TO_LIST = 'INSTALLED_APPS = ["random3", "random4", ]'
APPS_ASSIGNMENT_TO_LIST_MIXED = 'INSTALLED_APPS = ["random5", 3, ]'
TEST_RUNNER_ASSIGNMENT_TO_LIST = 'TEST_RUNNER = ["random6", "random7", ]'
TEST_RUNNER_ASSIGNMENT_TO_STRING = 'TEST_RUNNER = "random8"'


def _get_imports_of_python_module(folder, source):
    temporal_file = write_source_file_at(
        (folder.strpath,),
        source_code=source,
    )

    django_settings = DjangoSettings(folder.strpath, temporal_file)
    dotted_names = [x.name for x in django_settings.scan()]
    return dotted_names


def test_create_from_files_nothing(minimal_structure):
    path, package_name = minimal_structure
    modules_found = list(DjangoSettings.create_from_files(path))
    assert len(modules_found) == 0


def test_create_from_files_single_file_random_name():
    _, tmp_file = tempfile.mkstemp('.py')
    modules_found = list(DjangoSettings.create_from_files(tmp_file))
    assert len(modules_found) == 0


def test_create_from_files_single_file_settings_name(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    write_source_file_at([src_path], filename='some_settings.py')
    modules_found = list(DjangoSettings.create_from_files(src_path))
    assert len(modules_found) == 1


def test_create_from_files_deep_nested(minimal_structure):
    path, package_name = minimal_structure
    src_path = os.path.join(path, 'src')
    write_source_file_at(
        [src_path, 'a', 'b', 'c'],
        filename='anothersettings.py',
    )
    modules_found = list(DjangoSettings.create_from_files(src_path))
    assert len(modules_found) == 1


def test_random_code(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, RANDOM_CODE)
    assert len(dotted_names) == 0


def test_random_assignment(tmpdir):
    dotted_names = _get_imports_of_python_module(tmpdir, RANDOM_ASSIGNMENT)
    assert len(dotted_names) == 0


def test_apps_assignment_to_string(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_STRING,
    )
    assert len(dotted_names) == 0


def test_apps_assignment_to_tuple(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_TUPLE,
    )
    assert len(dotted_names) == 2


def test_apps_assignment_to_tuple_details(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_TUPLE,
    )
    assert 'random1' in dotted_names
    assert 'random2' in dotted_names


def test_apps_assignment_to_list(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_LIST,
    )
    assert len(dotted_names) == 2


def test_apps_assignment_to_list_details(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_LIST,
    )
    assert 'random3' in dotted_names
    assert 'random4' in dotted_names


def test_apps_assignment_to_list_mixed(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_LIST_MIXED,
    )
    assert len(dotted_names) == 1


def test_apps_assignment_to_list_mixed_details(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        APPS_ASSIGNMENT_TO_LIST_MIXED,
    )
    assert dotted_names[0] == 'random5'


def test_runner_assignment_to_list(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        TEST_RUNNER_ASSIGNMENT_TO_LIST,
    )
    assert len(dotted_names) == 0


def test_runner_assignment_to_string(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        TEST_RUNNER_ASSIGNMENT_TO_STRING,
    )
    assert len(dotted_names) == 1


def test_apps_assignment_to_string_mixed_details(tmpdir):
    dotted_names = _get_imports_of_python_module(
        tmpdir,
        TEST_RUNNER_ASSIGNMENT_TO_STRING,
    )
    assert dotted_names[0] == 'random8'



import os
import tempfile

import pytest

from z3c.dependencychecker.modules import BaseModule
from z3c.dependencychecker.tests.utils import write_source_file_at


def test_module_path():
    obj = BaseModule('/some/path', '/some/path/random/bla')
    assert obj.path == '/some/path/random/bla'


def test_scan_raises():
    obj = BaseModule('/some/path', '/some/path/random/bla')
    with pytest.raises(NotImplementedError):
        obj.scan()


def test_create_from_files_raises():
    with pytest.raises(NotImplementedError):
        BaseModule.create_from_files('/some/path')


def test_has_test_with_prefix_in_path():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, 'blatest'))
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_has_test_in_path():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, 'test'))
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_has_tests_in_path():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, 'tests'))
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_has_test_in_filename():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, 'bla'), filename='test.py')
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_has_tests_in_filename():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, 'bla'), filename='tests.py')
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_has_tests_with_suffix_in_filename():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, 'bla'), filename='testsohlala.py')
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing


def test_is_not_a_test_module():
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, 'bla'), filename='bla.py')
    python_module = BaseModule(folder, temporal_file)
    assert python_module.testing is False


def test_parent_folder_is_test_but_module_not():
    """Be sure to ignore the filesystem part and take into account the
    package part

    This means that if I have a package in a folder like:
    /home/gil/testing/my.package
    And a file within it like:
    /home/gil/testing/my.package/src/my/package/configure.zcml
    Despite having 'test' in the path, as it is not part of the package,
    it should be ignored.
    """
    folder = tempfile.mkdtemp()
    temporal_file = write_source_file_at((folder, 'test', 'bla'), filename='bla.py')
    python_module = BaseModule(os.path.join(folder, 'test'), temporal_file)
    assert python_module.testing is False



import os


class change_dir:
    """Step into a directory temporarily

    Copied from https://pythonadventures.wordpress.com/2013/12/15/chdir-a-context-manager-for-switching-working-directories/  # noqa: E501
    """

    def __init__(self, path):
        self.old_dir = os.getcwd()
        self.new_dir = path

    def __enter__(self):
        os.chdir(self.new_dir)

    def __exit__(self, *args):
        os.chdir(self.old_dir)



import sys
from unittest import mock

from pkg_resources import load_entry_point

from z3c.dependencychecker.main import main
from z3c.dependencychecker.utils import change_dir

MAIN_OUTPUT = """
Missing requirements
====================
     Products.GenericSetup.interfaces.EXTENSION
     missing.req
     other.generic.setup.dependency
     plone.app.content.interfaces.INameFromTitle
     plone.app.dexterity.behaviors.metadata.IBasic
     plone.random1.interfaces.IMySchema
     plone.random2.content.MyType
     some_django_app
     something.origname
     zope.exceptions
     zope.interface
     zope.interface.verify

Missing test requirements
=========================
     plone.dexterity.browser.views.ContentTypeView
     plone.dexterity.interfaces.IContentType
     reinout.hurray
     transaction
     zope.filerepresentation.interfaces.IRawReadFile

Unneeded requirements
=====================
     some.other.extension
     unneeded.req

Requirements that should be test requirements
=============================================
     Needed.By.Test

Unneeded test requirements
==========================
     zope.testing

Note: requirements are taken from the egginfo dir, so you need
to re-run buildout (or setup.py or whatever) for changes in
setup.py to have effect.

"""


def test_highlevel_integration(capsys, fake_project):
    with change_dir(fake_project):
        arguments = ['dependencychecker']
        try:
            with mock.patch.object(sys, 'argv', arguments):
                main()
        except SystemExit:
            out, err = capsys.readouterr()
            assert MAIN_OUTPUT in out
        else:
            assert True is False  # pragma: nocover


def test_entry_point_installed():
    """Check that pkg_resources can find the entry point defined in setup.py"""
    entry_point = load_entry_point(
        'z3c.dependencychecker', 'console_scripts', 'dependencychecker'
    )
    assert entry_point


def test_entry_point_run():
    """Check that calling the entry point calls a z3c.dependencychecker
    function
    """

    def fake_main():
        return 'All dependencies are fine'

    import z3c.dependencychecker.main

    with mock.patch.object(z3c.dependencychecker.main, 'main', fake_main):
        entry_point = load_entry_point(
            'z3c.dependencychecker', 'console_scripts', 'dependencychecker'
        )

    assert entry_point() == 'All dependencies are fine'
